#include <Wire.h>//biblioteka BME280 i AltIMU-10
#include <Adafruit_Sensor.h>//biblioteka BME280
#include <Adafruit_BME280.h>//biblioteka BME280

#include <LSM6.h>//biblioteka AltIMU-10 (akcelerometr i zyroskop)

#include <LIS3MDL.h>//biblioteka AltIMU-10 (magnetometr)

#include <LPS.h>//biblioteka AltIMU-10 (temperatura i cisnienie)

#include <SPI.h>//biblioteka komunikacji SPI

#include <SD.h>//biblioteka karty SD

#include <OneWire.h>//biblioteka DS18B20

#include <RH_RF95.h>//biblioteka nadajnika

#include <Adafruit_GPS.h> //biblioteka gps

#include <Arduino.h>   // required before wiring_private.h
#include "wiring_private.h" // pinPeripheral() function

Uart Serial2 (&sercom1, 11, 10, SERCOM_RX_PAD_0, UART_TX_PAD_2);

void SERCOM1_Handler()
{
  Serial2.IrqHandler();
}

#define GPSSerial Serial1

#define PMSSerial Serial2

Adafruit_BME280 bme;//zainicjowanie zmiennej do czujnika BME280

Adafruit_GPS GPS(&GPSSerial); //stworzenie obiektu GPS

File plik;//zainicjowanie zmiennej pliku tekstowego

uint32_t timer = millis();//utworzenie stopera do odmierzania czasu pomiedzy pomiarami

bool znalezione_SD = 0;

int pin_cs_nadajnik = 8;
int pin_RST_nadajnik = 4;
int pin_INT_nadajnik = 3;
#define czestotliwosc_nadajnika 433.0

RH_RF95 rf95(pin_cs_nadajnik, pin_INT_nadajnik);

int pin_cs_SD = 12;

OneWire  ds18b20(6);

LSM6 imu; //zmienna do pomiarow z AltIMU-10 (żyroskop i akcelerometr)

LIS3MDL mag; //zmienna do pomiarow z AltIMU-10 (magnetometr)

LPS ps; //zmienna do pomiarow z AltIMU-10 (temperatura i cisnienie)

//liczenie czasu potrzebnego kazdemu czujnikowi
uint32_t poczatek_calkowity;
uint32_t koniec_calkowity;
uint32_t poczatek;
uint32_t koniec;

//obiekt do pomiaru PMS7003
struct pms5003data
{
  uint16_t framelen;
  uint16_t pm10_standard, pm25_standard, pm100_standard;
  uint16_t pm10_env, pm25_env, pm100_env;
  uint16_t particles_03um, particles_05um, particles_10um, particles_25um, particles_50um, particles_100um;
  uint16_t unused;
  uint16_t checksum;
};

struct pms5003data data;


//BME280
/*
  VIN - 3.3V
  GND - GND
  SCL - SCL
  SDA - SDA
*/

//SD Card
/*
  wszystkie piny z jednej linii
  GND - GND
  VDD - 3.3V
  DI - MOSI
  DO - MISO
  SCLK - SCK
  CS - pin 12
*/

//LoRa Radio
/*
  VCC - 3.3V
  GND - GND
  MOSI - MOSI
  MISO - MISO
  CLK - SCK
  CS - pin 8
  RST - pin 4
  INT - pin 3
*/

//DS18B20
/*
  numeracja od lewej strony (płaska)
  1 pin - GND
  2 pin - pin 6 + 3.3V przez rezystor 4.7k Ohm
  3 pin - 3.3V
*/

//GPS
/*
  VCC - 3.3V
  GND - GND
  TX - RX
  RX - TX
*/

//Akcelerometr i zyroskop
/*
  VIN - 3.3V
  GND - GND
  SCL - SCL
  SDA - SDA
*/

//PMS
/*
  VCC - Step-Up VOUT - 3.3V
  GND - GND
  RX - 10
  TX - 11
*/

//Step-UP
/*
  VIN - 3.3V
  GND - GND
  VOUT - napiecie wyjsciowe 5V
*/

void mierzenie_napiecia_baterii(double &napiecie_baterii)
{
  napiecie_baterii = analogRead(A7) * 6.6 / 1024;
}

void odczytywanie_BME280(double &temperatura, double &cisnienie_BME280, double &wilgotnosc)
{
  bme.read(temperatura, cisnienie_BME280, wilgotnosc);
}

void zapisywanie_SD(double temperatura_BME280, double cisnienie_BME280, double wilgotnosc, double temperatura_DS18B20, int godzina, int minuta, int sekunda, int dzien, int miesiac, int rok, double lokalizacja[], char lokalizacja_znak[], double szybkosc_wezly, double kat, double wysokosc, int liczba_satelit, double akcelerometr[], double zyroskop[], double magnetometr[], double temperatura_altimu, double cisnienie_altimu, int PM1, int PM2_5, int PM10, double napiecie_baterii)
{
  digitalWrite(pin_cs_SD, LOW);
  plik = SD.open("test.txt", FILE_WRITE);
  if (plik)
  {
    znalezione_SD = 1;
    Serial.println("Zapis SD");
    plik.print(temperatura_BME280);
    plik.print(";");
    plik.print(cisnienie_BME280);
    plik.print(";");
    plik.print(wilgotnosc);
    plik.print(";");
    plik.print(temperatura_DS18B20);
    plik.print(";");
    plik.print(godzina);
    plik.print(";");
    plik.print(minuta);
    plik.print(";");
    plik.print(sekunda);
    plik.print(";");
    plik.print(dzien);
    plik.print(".");
    plik.print(miesiac);
    plik.print(".");
    plik.print(rok);
    plik.print(";");
    plik.print(lokalizacja[0], 4);
    plik.print(";");
    plik.print(lokalizacja_znak[0]);
    plik.print(";");
    plik.print(lokalizacja[1], 4);
    plik.print(";");
    plik.print(lokalizacja_znak[1]);
    plik.print(";");
    plik.print(szybkosc_wezly);
    plik.print(";");
    plik.print(kat);
    plik.print(";");
    plik.print(wysokosc);
    plik.print(";");
    plik.print(liczba_satelit);
    plik.print(";");
    plik.print(akcelerometr[0]);
    plik.print(";");
    plik.print(akcelerometr[1]);
    plik.print(";");
    plik.print(akcelerometr[2]);
    plik.print(";");
    plik.print(zyroskop[0]);
    plik.print(";");
    plik.print(zyroskop[1]);
    plik.print(";");
    plik.print(zyroskop[2]);
    plik.print(";");
    plik.print(magnetometr[0]);
    plik.print(";");
    plik.print(magnetometr[1]);
    plik.print(";");
    plik.print(magnetometr[2]);
    plik.print(";");
    plik.print(temperatura_altimu);
    plik.print(";");
    plik.print(cisnienie_altimu);
    plik.print(";");
    plik.print(PM1);
    plik.print(";");
    plik.print(PM2_5);
    plik.print(";");
    plik.print(PM10);
    plik.print(";");
    plik.println(napiecie_baterii);
  }
  else
  {
    Serial.println("Nie mozna otworzyc pliku na karcie SD");
    znalezione_SD = 0;
  }

  plik.close();
  digitalWrite(pin_cs_SD, HIGH);
}

void odczytywanie_DS18B20(double &temperatura_DS18B20)
{
  byte present = 0;
  byte type_s;
  byte data[12];
  byte addr[8];
  double celsius;

  ds18b20.reset_search();
  if (!ds18b20.search(addr))
  {
    ds18b20.reset_search();
  }

  if (OneWire::crc8(addr, 7) == addr[7])
  {
    ds18b20.reset();
    ds18b20.select(addr);
    ds18b20.write(0x44, 1);

    present = ds18b20.reset();
    ds18b20.select(addr);
    ds18b20.write(0xBE);

    for (byte i = 0; i < 9; i++)
    {
      data[i] = ds18b20.read();
    }

    int16_t raw = (data[1] << 8) | data[0];

    if (addr[0] == 0x28)
    {
      byte cfg = (data[4] & 0x60);
      if (cfg == 0x00)
        raw = raw & ~7;
      else if (cfg == 0x20)
        raw = raw & ~3;
      else if (cfg == 0x40)
        raw = raw & ~1;

      celsius = (double)raw / 16.0;
      temperatura_DS18B20 = celsius;
    }
  }
}

void odczytywanie_stale_gps()
{
  GPS.read();

  if (GPS.newNMEAreceived())
  {
    if (!GPS.parse(GPS.lastNMEA()))
    {
      return;
    }
  }
}

void odczytywanie_gps(int &godzina, int &minuta, int &sekunda, int &dzien, int &miesiac, int &rok, int &fix, int &quality, double lokalizacja[], char lokalizacja_znak[], double &szybkosc_wezly, double &kat, double &wysokosc, int &liczba_satelit)
{

  //CZAS
  godzina = GPS.hour;
  minuta = GPS.minute;
  sekunda = GPS.seconds;

  //DATA
  dzien = GPS.day;
  miesiac = GPS.month;
  rok = GPS.year;

  fix = (int)GPS.fix;
  quality = (int)GPS.fixquality;

  if (fix)
  {
    lokalizacja[0] = GPS.latitude;
    lokalizacja_znak[0] = GPS.lat;
    lokalizacja[1] = GPS.longitude;
    lokalizacja_znak[1] = GPS.lon;
    szybkosc_wezly = GPS.speed;
    kat = GPS.angle;
    wysokosc = GPS.altitude;
    liczba_satelit = (int)GPS.satellites;
  }
}

boolean readPMSdata(Stream *s)
{
  if (! s->available())
  {
    return false;
  }

  //odczytuje bit i szuka startowego '0x42'
  if (s->peek() != 0x42)
  {
    s->read();
    return false;
  }

  //Odczytuje 32 bity
  if (s->available() < 32)
  {
    return false;
  }

  uint8_t buffer[32];
  uint16_t sum = 0;
  s->readBytes(buffer, 32);


  for (uint8_t i = 0; i < 30; i++)
  {
    sum += buffer[i];
  }


  uint16_t buffer_u16[15];
  for (uint8_t i = 0; i < 15; i++)
  {
    buffer_u16[i] = buffer[2 + i * 2 + 1];
    buffer_u16[i] += (buffer[2 + i * 2] << 8);
  }

  //umieszcza dane w strukturze "data"
  memcpy((void *)&data, (void *)buffer_u16, 30);

  if (sum != data.checksum)
  {
    return false;
  }
  return true;
}

void odczytywanie_PMS7003(int &PM1, int &PM2_5, int &PM10)
{
  
  for (int i = 0; i < 40; i++)
  {
    if (readPMSdata(&PMSSerial))
    {
      PM1 = data.pm10_standard;
      PM2_5 = data.pm25_standard;
      PM10 = data.pm100_standard;
      i = 40;
    }
  }
}

void odczytywanie_altimu(double akcelerometr[], double zyroskop[], double magnetometr[], double &temperatura_altimu, double &cisnienie_altimu)
{
  imu.read();
  
  akcelerometr[0] = imu.a.x * 0.488 / 1000;
  akcelerometr[1] = imu.a.y * 0.488 / 1000;
  akcelerometr[2] = imu.a.z * 0.488 / 1000;
  zyroskop[0] = imu.g.x * 70 / 1000.0;
  zyroskop[1] = imu.g.y * 70 / 1000.0;
  zyroskop[2] = imu.g.z * 70 / 1000.0;
  
  mag.read();

  magnetometr[0] = mag.m.x / 6842.0;
  magnetometr[1] = mag.m.y / 6842.0;
  magnetometr[2] = mag.m.z / 6842.0;
  
  ps.read(temperatura_altimu, cisnienie_altimu);
}

void wyswietlanie(double temperatura_BME280, double cisnienie_BME280, double wilgotnosc, double temperatura_DS18B20, int godzina, int minuta, int sekunda, int dzien, int miesiac, int rok, int fix, int quality, double lokalizacja[], char lokalizacja_znak[], double szybkosc_wezly, double kat, double wysokosc, int liczba_satelit, double akcelerometr[], double zyroskop[], double magnetometr[], double temperatura_altimu, double cisnienie_altimu, int PM1, int PM2_5, int PM10, double napiecie_baterii)
{
  Serial.print("Temperatura_BME280: ");
  Serial.print(temperatura_BME280);
  Serial.print("*C cisnienie_BME280: ");
  Serial.print(cisnienie_BME280);
  Serial.print("hPa Wilgotnosc: ");
  Serial.print(wilgotnosc);
  Serial.print("% Temperatura_DS18B20: ");
  Serial.print(temperatura_DS18B20);
  Serial.println("*C");
  Serial.print("Czas: ");
  Serial.print(godzina);
  Serial.print(":");
  Serial.print(minuta);
  Serial.print(":");
  Serial.print(sekunda);
  Serial.print(" Data: ");
  Serial.print(dzien);
  Serial.print(".");
  Serial.print(miesiac);
  Serial.print(".");
  Serial.print(rok);
  Serial.print(" Fix: ");
  Serial.print(fix);
  Serial.print(" Quality: ");
  Serial.print(quality);
  Serial.print(" Lokalizacja: ");
  Serial.print(lokalizacja[0], 4);
  Serial.print(lokalizacja_znak[0]);
  Serial.print(" ");
  Serial.print(lokalizacja[1], 4);
  Serial.println(lokalizacja_znak[1]);
  Serial.print("Szybkosc (wezly): ");
  Serial.print(szybkosc_wezly);
  Serial.print(" Kat: ");
  Serial.print(kat);
  Serial.print(" Wysokosc: ");
  Serial.print(wysokosc);
  Serial.print(" Liczba satelit: ");
  Serial.println(liczba_satelit);
  Serial.print("Akcelerometr: ");
  Serial.print(akcelerometr[0]);
  Serial.print(" ");
  Serial.print(akcelerometr[1]);
  Serial.print(" ");
  Serial.print(akcelerometr[2]);
  Serial.print(" Zyroskop: ");
  Serial.print(zyroskop[0]);
  Serial.print(" ");
  Serial.print(zyroskop[1]);
  Serial.print(" ");
  Serial.print(zyroskop[2]);
  Serial.print(" Magnetometr: ");
  Serial.print(magnetometr[0]);
  Serial.print(" ");
  Serial.print(magnetometr[1]);
  Serial.print(" ");
  Serial.println(magnetometr[2]);
  Serial.print("Temperatura_AltIMU: ");
  Serial.print(temperatura_altimu);
  Serial.print("*C Cisnienie_AltIMU: ");
  Serial.print(cisnienie_altimu);
  Serial.print("hPa PM1: ");
  Serial.print(PM1);
  Serial.print(" PM2.5: ");
  Serial.print(PM2_5);
  Serial.print(" PM10: ");
  Serial.print(PM10);
  Serial.print(" Stan baterii: ");
  Serial.println(napiecie_baterii);
}


void setup()
{
  Serial.begin(115200);
  Serial.println("start");

  PMSSerial.begin(9600);

  pinPeripheral(10, PIO_SERCOM);
  pinPeripheral(11, PIO_SERCOM);

  Serial.println("Inicjalizacja biblioteki Wire");
  Wire.begin(); //zainicjowanie biblioteki wire

  Serial.println("Inicjalizacja AltIMU (zyroskop i akcelerometr)");
  imu.init();//zainicjowanie AltIMU-10 (zyroskop i akcelerometr)
  imu.enableDefault();//zainicjowanie AltIMU-10 (zyroskop i akcelerometr)

  Serial.println("Inicjalizacja AltIMU (magnetometr)");
  mag.init();//zainicjowanie AltIMU-10 (magnetometr)
  mag.enableDefault();//zainicjowanie AltIMU-10 (magnetometr)

  Serial.println("Inicjalizacja AltIMU (temperatura i cisnienie)");
  ps.init();//zainicjowanie AltIMU-10 (temperatura i cisnienie)
  ps.enableDefault();//zainicjowanie AltIMU-10 (temperatura i cisnienie)

  GPS.begin(9600);
  GPS.sendCommand(PMTK_SET_NMEA_OUTPUT_RMCGGA);
  GPS.sendCommand(PMTK_SET_NMEA_UPDATE_1HZ);
  GPS.sendCommand(PGCMD_ANTENNA);
  delay(1000);
  GPSSerial.println(PMTK_Q_RELEASE);

  SPI.begin(); //inicjowanie komunikacji SPI

  Serial.println("Inicjalizacja BME280");
  bme.begin(0x76); //inicjowanie BME280


  pinMode(pin_cs_SD, OUTPUT);
  digitalWrite(pin_cs_SD, LOW);

  if (!SD.begin(pin_cs_SD))
  {
    Serial.println("Nie znaleziono karty SD");
    znalezione_SD = 0;
  }
  else
  {
    Serial.println("Znaleziono karte SD");
    znalezione_SD = 1;
  }
  digitalWrite(pin_cs_SD, HIGH);
}

void loop()
{
  odczytywanie_stale_gps();

  if (millis() - timer > 1000)
  {
    digitalWrite(13, HIGH);
    
    poczatek_calkowity = millis();
    timer = millis(); // reset the timer

    double temperatura_BME280 = 0;///////
    double cisnienie_BME280 = 0;
    double wilgotnosc = 0;///////////
    double temperatura_DS18B20 = 0;
    int godzina = 0;
    int minuta = 0;
    int sekunda = 0;
    int dzien = 0;
    int miesiac = 0;
    int rok = 0;
    int fix = 0;
    int quality = 0;
    double lokalizacja[2] = {0, 0};
    char lokalizacja_znak[2] = {' ', ' '};
    double szybkosc_wezly = 0;
    double kat = 0;
    double wysokosc = 0;
    int liczba_satelit = 0;
    double akcelerometr[3] = {0, 0, 0};
    double zyroskop[3] = {0, 0, 0};
    double magnetometr[3] = {0, 0, 0};/////////
    double temperatura_altimu = 0;////////
    double cisnienie_altimu = 0;//////////
    int PM1 = 0;
    int PM2_5 = 0;
    int PM10 = 0;

    double napiecie_baterii = 0;

    poczatek = millis();
    odczytywanie_gps(godzina, minuta, sekunda, dzien, miesiac, rok, fix, quality, lokalizacja, lokalizacja_znak, szybkosc_wezly, kat, wysokosc, liczba_satelit);
    koniec = millis();
    Serial.print("Czas wykonania GPS: ");
    Serial.println(koniec - poczatek);

    poczatek = millis();
    odczytywanie_BME280(temperatura_BME280, cisnienie_BME280, wilgotnosc);
    koniec = millis();
    Serial.print("Czas wykonania BME280: ");
    Serial.println(koniec - poczatek);

    poczatek = millis();
    odczytywanie_DS18B20(temperatura_DS18B20);
    koniec = millis();
    Serial.print("Czas wykonania DS18B20: ");
    Serial.println(koniec - poczatek);

    poczatek = millis();
    odczytywanie_altimu(akcelerometr, zyroskop, magnetometr, temperatura_altimu, cisnienie_altimu);
    koniec = millis();
    Serial.print("Czas wykonania AltIMU: ");
    Serial.println(koniec - poczatek);

    poczatek = millis();
    odczytywanie_PMS7003(PM1, PM2_5, PM10);
    koniec = millis();
    Serial.print("Czas wykonania PMS7003: ");
    Serial.println(koniec - poczatek);

    mierzenie_napiecia_baterii(napiecie_baterii);
    
    poczatek = millis();
    wyswietlanie(temperatura_BME280, cisnienie_BME280, wilgotnosc, temperatura_DS18B20, godzina, minuta, sekunda, dzien, miesiac, rok, fix, quality, lokalizacja, lokalizacja_znak, szybkosc_wezly, kat, wysokosc, liczba_satelit, akcelerometr, zyroskop, magnetometr, temperatura_altimu, cisnienie_altimu, PM1, PM2_5, PM10, napiecie_baterii);
    koniec = millis();
    Serial.print("Czas wykonania wyswietlanie: ");
    Serial.println(koniec - poczatek);

    poczatek = millis();
    if(znalezione_SD == 1)
    {
      zapisywanie_SD(temperatura_BME280, cisnienie_BME280, wilgotnosc, temperatura_DS18B20, godzina, minuta, sekunda, dzien, miesiac, rok, lokalizacja, lokalizacja_znak, szybkosc_wezly, kat, wysokosc, liczba_satelit, akcelerometr, zyroskop, magnetometr, temperatura_altimu, cisnienie_altimu, PM1, PM2_5, PM10, napiecie_baterii);
    }
    koniec = millis();
    Serial.print("Czas wykonania SD: ");
    Serial.println(koniec - poczatek);


    koniec_calkowity = millis();
    Serial.print("Czas wykonania programu: ");
    Serial.println(koniec_calkowity - poczatek_calkowity);

    digitalWrite(13, LOW);
  }
}
